
#pragma once

struct fsm_element;
typedef boost::intrusive_ptr<fsm_element> fsm_element_ptr;

struct fsm_element
{
   fsm_element(combinator const & my)
      : my_(my)
      , ref_counter_(0)
   {}

   void add_branch_to(unsigned id, fsm_element_ptr target)
   {
      if (id < branches_.size())
      {
         if (branches_[id])
            throw std::logic_error("overwrite branches");
      }
      else
      {
         branches_.resize(id + 1);
      }

      branches_[id] = target;
   }

   fsm_element_ptr get_branch(unsigned id) const
   {
      if (id >= branches_.size())
         return 0;

      return branches_[id];
   }

   bool check_for_apply(element_list & e)
   {
      return my_.check_for_apply(e);
   }

   void apply(element_list & e)
   {
      my_.apply(e);
   }

private:
   combinator my_;
   std::vector<fsm_element_ptr> branches_;

private:
   unsigned ref_counter_;

   friend void intrusive_ptr_add_ref(fsm_element * p)
   {
      ++p->ref_counter_;
   }

   friend void intrusive_ptr_release(fsm_element * p)
   {
      if (--p->ref_counter_ == 0)
         delete p;
   }
};

inline combinator create_null_combinator()
{
   return combinator();
}

inline combinator create_i_combinator()
{
   return combinator(1, element::argument(0));
}

inline combinator create_k_combinator()
{
   return combinator(2, element::argument(0));
}

inline combinator create_s_combinator()
{
   element_list * pattern = new element_list();
   element pe(element::ref(pattern));

   element_list * pattern2 = new element_list();
   element pe2(element::ref(pattern2));
   pattern2->push(element::argument(2));
   pattern2->push(element::argument(1));

   pattern->push(pe2);
   pattern->push(element::argument(2));
   pattern->push(element::argument(0));

   return combinator(3, pe);
}

struct fsm
{
   fsm()
      : root_(new fsm_element(create_null_combinator()))
   {
      cte_ctx_.i = add_root_branch(create_i_combinator());
      cte_ctx_.k = add_root_branch(create_k_combinator());
      cte_ctx_.s = add_root_branch(create_s_combinator());
   }

   fsm_element_ptr get_root()
   {
      return root_;
   }

   convert_to_element_context get_cte_context() const
   {
      return cte_ctx_;
   }

   combinator_list & get_combinator_list()
   {
      return cl_;
   }

private:
   unsigned add_root_branch(combinator const & c)
   {
      size_t id = cl_.add(c);
      root_->add_branch_to(id, new fsm_element(c));
      return id;
   }

private:
   combinator_list cl_;
   fsm_element_ptr root_;
   convert_to_element_context cte_ctx_;
};

inline bool try_apply(element_list & el, fsm_element_ptr current, std::vector<element> & fetched)
{
   if (!current->check_for_apply(el))
   {
      while (!fetched.empty())
      {
         el.push(fetched.back());
         fetched.pop_back();
      }
      return false;
   }
   else
   {
      current->apply(el);
      return true;
   }
}

inline bool try_apply_smart(element_list & el, fsm_element_ptr current, std::vector<element> & fetched, fsm & fsm)
{
   element_list * sand_box = new element_list();
   element sand_box_hold = element::ref(sand_box);

   assert(el.top().is_combinator());
   unsigned new_branch = el.top().as_combinator();
   sand_box->push(el.top());
   el.pop();

   while (!fetched.empty())
   {
      sand_box->push(fetched.back());
      fetched.pop_back();
   }

   std::wcout << L"generate_combinator: ";
   el2s_rec(std::wcout, *sand_box);
   std::wcout << std::endl;

   combinator new_combinator = generate_combinator(fsm.get_combinator_list(), *sand_box);
   unsigned new_combinator_id = fsm.get_combinator_list().add(new_combinator);

   fsm_element_ptr new_fsm_element = new fsm_element(new_combinator);
   current->add_branch_to(new_branch, new_fsm_element);
   fsm.get_root()->add_branch_to(new_combinator_id, new_fsm_element); // TODO: add other paths
   if (!new_fsm_element->check_for_apply(el))
   {
      el.push(element::combinator(new_combinator_id));
      return false;
   }
   else
   {
      new_fsm_element->apply(el);
      return true;
   }
}

struct fsm_evaluator
{
   fsm_evaluator(fsm & fsm)
      : fsm_(fsm)
      , reduction_count_(0)
   {}

   bool one_step_reduce(element_list & el)
   {
//       std::wstringstream wss;
//       wss << L"one_step_reduce: ";
//       el2c(wss, el);
//       wss << std::endl;

      element t = el.top();
      if (t.is_argument())
         return false;
      if (t.is_ref())
      {
         element_list * nested = el.top().as_ref();
         full_reduce(*nested);

         element hold = el.top();
         el.pop();
         for (element_list::const_iterator i = nested->begin(); i != nested->end(); ++i)
            el.push(*i);

         return true;
      }

      if (!t.is_combinator())
         throw std::logic_error("bad element type");

//       std::wcout << L"one_step_reduce*: ";
//       el2s_rec(std::wcout, el);
//       std::wcout << std::endl;

      std::vector<element> fetched;
      fsm_element_ptr current = fsm_.get_root()->get_branch(t.as_combinator());
      fetched.push_back(t);
      el.pop();

      for (;;)
      {
         if (current == 0)
            throw std::logic_error("bad fsm");

         if (el.empty() || el.top().is_ref() || el.top().is_argument())
         {
            bool r = try_apply(el, current, fetched);
            if (r)
            {
               ++reduction_count_;
//               std::wcout << wss.str();
            }
            return r;
         }

         fsm_element_ptr next = current->get_branch(el.top().as_combinator());
         if (!next)
         {
//            bool r = try_apply(el, current, fetched);
            bool r = try_apply_smart(el, current, fetched, fsm_);
            if (r)
            {
               ++reduction_count_;
//               std::wcout << wss.str();
            }
            return r;
         }

         current = next;
         fetched.push_back(el.top());
         el.pop();
      }
   }

   void full_reduce(element_list & el)
   {
//      std::wcout << L"--- full_reduce ---" << std::endl;
      while (one_step_reduce(el));
//      std::wcout << L"--- end full reduce ---" << std::endl;
   }

   void completely_full_reduce(element_list & el)
   {
      full_reduce(el);
      for (element_list::reverse_iterator i = boost::next(el.rbegin()); i != el.rend(); ++i)
      {
         if (i->is_ref())
         {
            element_list * nested = i->as_ref();
            completely_full_reduce(*nested);
            if (nested->size() == 1)
               *i = nested->top();
         }
      }
   }

   convert_to_element_context get_cte_context() const
   {
      return fsm_.get_cte_context();
   }

   unsigned get_reduction_count() const
   {
      return reduction_count_;
   }

private:
   fsm & fsm_;
   unsigned reduction_count_;
};